import logging
import time
from typing import Text, List

import jwt
from sanic import Blueprint, response
from sanic.request import Request

from rasa.core import events
from rasa.core.trackers import EventVerbosity
from rasa.core.training.dsl import StoryParseError
from rasa.core.training.structures import Story
from rasax.community import utils, config
from rasax.community.api.decorators import platform_scoped, inject_rasa_x_user
from rasax.community.services.event_service import EventService, FLAGGED_MESSAGES_KEY
from rasax.community.services.projects_service import ProjectsService
from rasax.community.services.role_service import RoleService
from rasax.community.services.role_service import (
    normalise_permissions,
    GUEST_ROLE,
    guest_permissions,
)
from rasax.community.services.settings_service import SettingsService
from rasax.community.services.stack_service import StackService
from rasax.community.services.story_service import StoryService
from rasax.community.services.user_service import has_role
from rasax.community.utils import error, convert_bytes_to_string

logger = logging.getLogger(__name__)


def _event_verbosity_from_request(request):
    if utils.bool_arg(request, "history", default=True):
        return EventVerbosity.ALL
    else:
        return EventVerbosity.AFTER_RESTART


def flag_event(event, is_flagged):
    tmp = event.copy()
    tmp["is_flagged"] = is_flagged
    return tmp


def _story_service(request: Request) -> StoryService:
    return StoryService(request["db_session"])


def _project_service(request: Request) -> ProjectsService:
    return ProjectsService(request["db_session"])


def _stack_service(request: Request, default_environment: Text) -> StackService:
    settings_service = SettingsService(request["db_session"])

    environment = utils.deployment_environment_from_request(
        request, default_environment
    )
    service = settings_service.get_stack_service(environment)
    if not service:
        raise ValueError(
            "Service for requested environment '{}' not found".format(environment)
        )

    return service


def _event_service(request: Request) -> EventService:
    return EventService(request["db_session"])


def _role_service(request: Request) -> RoleService:
    return RoleService(request["db_session"])


def truncate_float(_float: float, decimal_places: int = 4) -> float:
    """Truncate float to `decimal_places` after the decimal separator."""

    return float("%.{}f".format(decimal_places) % _float)


async def get_truncated_timestamps(
    request: Request, conversation_id: Text, decimal_places: int = 4
) -> List[float]:
    """Create a list of flagged timestamps for `conversation_id` truncated to
        `decimal_places` after the decimal separator."""

    flagged_timestamps = _event_service(request).get_flagged_message_timestamps(
        conversation_id
    )
    return [
        truncate_float(timestamp, decimal_places=decimal_places)
        for timestamp in flagged_timestamps
    ]


def blueprint() -> Blueprint:
    stack_endpoints = Blueprint("stack_endpoints")

    @stack_endpoints.route("/conversations", methods=["GET", "HEAD"])
    @platform_scoped("metadata.get", allow_api_token=True)
    @inject_rasa_x_user()
    async def list_clients(request, user=None):
        clients, number_clients = _event_service(request).get_clients(request, user)
        return response.json(clients, headers={"X-Total-Count": number_clients})

    @stack_endpoints.route("/conversations", methods=["POST"])
    @platform_scoped("metadata.create", allow_api_token=True)
    async def create_conversation(request):
        conversation = request.json
        if not utils.check_schema("conversation", conversation):
            return error(
                400,
                "WrongSchema",
                "Please check the schema of your conversation payload. "
                "See https://rasa.com/docs/platform/api/conversations"
                "/#operation/createConversation for the correct format.",
            )

        try:
            inserted = _event_service(request).create_conversation_from(conversation)

            # initialise tracker (will just contain action_listen)
            await _stack_service(
                request, "production"
            ).initialise_tracker_for_conversation(inserted["sender_id"])

            return response.json(inserted, 201)
        except ValueError as e:
            return error(400, "SenderIdExists", details=e)

    @stack_endpoints.route("/conversations/<conversation_id>", methods=["GET", "HEAD"])
    @platform_scoped("clients.get", allow_api_token=True)
    @inject_rasa_x_user(allow_api_token=True, extract_user_from_jwt=True)
    async def get_events(request, conversation_id, user=None):
        if not has_role(user, "admin") and conversation_id != user["username"]:
            return error(401, "NoPermission", "Access denied")

        until_time = utils.float_arg(request, "until", None)
        event_verbosity = _event_verbosity_from_request(request)

        try:
            tracker = await _stack_service(request, "production").tracker_json(
                conversation_id, event_verbosity, until=until_time
            )
        except ValueError as e:
            return error(404, "ServiceNotFound", details=e)

        if tracker is not None:
            accept = request.headers.get("Accept")

            flagged_timestamps = await get_truncated_timestamps(
                request, conversation_id
            )

            tracker[FLAGGED_MESSAGES_KEY] = flagged_timestamps

            for i, e in enumerate(tracker["events"]):
                event_timestamp = truncate_float(
                    float(e.get("timestamp", -1)), decimal_places=4
                )
                is_flagged = event_timestamp in flagged_timestamps
                tracker["events"][i] = flag_event(e, is_flagged)

            if accept == "application/json":
                dispo = "attachment;filename={}-dump.json".format(conversation_id)
                return response.json(
                    tracker,
                    content_type="application/json",
                    headers={"Content-Disposition": dispo},
                )
            elif accept == "text/markdown":
                evts = events.deserialise_events(tracker["events"])
                story = Story.from_events(evts)
                exported = story.as_story_string(flat=True)
                return response.text(
                    exported,
                    content_type="text/markdown",
                    headers={
                        "Content-Disposition": "attachment;filename={}-story.md"
                        "".format(conversation_id)
                    },
                )
            else:
                return response.json(tracker)
        else:
            return error(
                404,
                "ClientNotFound",
                "Client for conversation_id {} could not be found"
                "".format(conversation_id),
            )

    @stack_endpoints.route("/conversationActions", methods=["GET", "HEAD"])
    @platform_scoped("conversationActions.list", allow_api_token=True)
    async def unique_actions(request):
        actions = _event_service(request).get_unique_actions()
        return response.json(actions, headers={"X-Total-Count": len(actions)})

    @stack_endpoints.route("/conversationIntents", methods=["GET", "HEAD"])
    @platform_scoped("conversationIntents.list", allow_api_token=True)
    async def unique_intents(request):
        intents = _event_service(request).get_unique_intents()
        return response.json(intents, headers={"X-Total-Count": len(intents)})

    @stack_endpoints.route("/conversationEntities", methods=["GET", "HEAD"])
    @platform_scoped("conversationEntities.list", allow_api_token=True)
    async def unique_entities(request):
        entities = _event_service(request).get_unique_entities()
        return response.json(entities, headers={"X-Total-Count": len(entities)})

    @stack_endpoints.route("/conversationPolicies", methods=["GET", "HEAD"])
    @platform_scoped("conversationPolicies.list", allow_api_token=True)
    async def unique_policies(request):
        policies = _event_service(request).get_unique_policies()
        return response.json(policies, headers={"X-Total-Count": len(policies)})

    @stack_endpoints.route("/conversations/<conversation_id>/events", methods=["POST"])
    @platform_scoped("clientEvents.create", allow_api_token=True)
    async def post_event(request, conversation_id):
        _event = request.json
        try:
            result, status = await _stack_service(
                request, "production"
            ).append_event_to_tracker(conversation_id, _event)

            if status == 200:
                return response.json(_event)
            else:
                logger.warning(
                    "Creating event for sender '{}' failed with status '{}' "
                    "and error '{}'.".format(conversation_id, status, result)
                )
                return error(
                    status,
                    "EventCreationError",
                    "Error when creating event for sender '{}'."
                    "".format(conversation_id),
                )
        except ValueError as e:
            return error(404, "ServiceNotFound", details=e)

    @stack_endpoints.route("/conversations/<conversation_id>/events", methods=["PUT"])
    @platform_scoped("clientEvents.update", allow_api_token=True)
    async def update_events(request, conversation_id):
        _events = request.json
        result, status = await _stack_service(request, "production").update_events(
            conversation_id, _events, _event_service(request)
        )

        if status == 200:
            return response.json(_events)
        else:
            logger.warning(
                "Updating events for sender '{}' failed with status '{}' "
                "and error '{}'.".format(conversation_id, status, result)
            )
            return error(
                status,
                "EventUpdatingError",
                "Error when updating events for sender '{}'."
                "".format(conversation_id),
            )

    @stack_endpoints.route("/conversations/<conversation_id>/execute", methods=["POST"])
    @platform_scoped("actionPrediction.create", allow_api_token=True)
    async def execute_action(request, conversation_id):
        action = request.json
        event_verbosity = _event_verbosity_from_request(request)

        result, status = await _stack_service(request, "production").execute_action(
            conversation_id, action, event_verbosity
        )

        if status == 200:
            return response.json(result)
        else:
            logger.warning(
                "Executing action for sender '{}' failed with status '{}' "
                "and error '{}'.".format(conversation_id, status, result)
            )
            return error(
                status,
                "ActionExecutionFailed",
                "Error when executing action for sender '{}'."
                "".format(conversation_id),
            )

    @stack_endpoints.route(
        "/conversations/<conversation_id>/messages", methods=["GET", "HEAD"]
    )
    @platform_scoped("clientMessages.list")
    @inject_rasa_x_user()
    async def get_messages(request, conversation_id, user=None):
        if not has_role(user, "admin") and conversation_id != user["username"]:
            return error(401, "NoPermission", "Access denied")

        until_time = utils.float_arg(request, "until", None)

        event_verbosity = _event_verbosity_from_request(request)

        try:
            tracker = await _stack_service(request, "production").tracker_json(
                conversation_id, event_verbosity, until=until_time
            )
        except ValueError as e:
            return error(404, "ServiceNotFound", details=e)

        if tracker is not None:
            messages = []
            flagged_timestamps = await get_truncated_timestamps(
                request, conversation_id
            )

            for i, e in enumerate(tracker["events"]):
                event_timestamp = truncate_float(
                    float(e.get("timestamp", -1)), decimal_places=4
                )
                is_flagged = event_timestamp in flagged_timestamps
                tracker["events"][i] = flag_event(e, is_flagged)

                if e.get("event") in {"user", "bot", "agent"}:
                    m = e.copy()
                    m["data"] = m.get("data", m.get("parse_data", None))
                    m["type"] = m["event"]
                    del m["event"]
                    m["is_flagged"] = is_flagged
                    messages.append(m)
            tracker["messages"] = messages
            return response.json(tracker, headers={"X-Total-Count": len(messages)})
        else:
            return error(
                404,
                "ClientNotFound",
                "Client for conversation_id {} could not be found"
                "".format(conversation_id),
            )

    @stack_endpoints.route(
        "/conversations/<conversation_id>/messages", methods=["POST"]
    )
    @platform_scoped("messages.create")
    async def send_message(request, conversation_id, user=None):
        return await chat_to_bot(request)

    @stack_endpoints.route("/statistics", methods=["GET", "HEAD"])
    @platform_scoped("statistics.get")
    @inject_rasa_x_user()
    async def get_statistics(request, user=None):
        event_service = _event_service(request)
        if not has_role(user, "admin"):
            stats = event_service.user_statistics_dict()
        else:
            stats = event_service.get_statistics()
        return response.json(stats)

    @stack_endpoints.route("/stories", methods=["GET", "HEAD"])
    @platform_scoped("stories.list", allow_api_token=True)
    async def get_stories(request):
        from rasa.core.domain import Domain

        text_query = utils.default_arg(request, "q", None)
        fields = utils.fields_arg(request, {"name", "annotation.user", "id"})
        id_query = utils.list_arg(request, "id")

        distinct = utils.bool_arg(request, "distinct", True)
        stories = _story_service(request).fetch_stories(
            text_query, fields, id_query=id_query, distinct=distinct
        )

        content_type = request.headers.get("Accept")
        if content_type == "text/vnd.graphviz":
            project_id = utils.default_arg(request, "project_id", config.project_name)
            domain_dict = _project_service(request).get_domain(project_id)
            domain = Domain.from_dict(domain_dict)

            visualization = await _story_service(request).visualize_stories(
                stories, domain
            )

            if visualization:
                return response.text(visualization)
            else:
                return error(
                    406,
                    "VisualizationNotAvailable",
                    "Cannot produce a visualization for the requested stories",
                )
        elif content_type == "text/markdown":
            markdown = _story_service(request).get_stories_markdown(stories)
            return response.text(
                markdown,
                content_type="text/markdown",
                headers={"Content-Disposition": "attachment;filename=stories.md"},
            )
        else:
            return response.json(stories, headers={"X-Total-Count": len(stories)})

    @stack_endpoints.route("/stories/<story_id>", methods=["GET", "HEAD"])
    @platform_scoped("stories.get", allow_api_token=True)
    async def get_story(request, story_id):
        from rasa.core.domain import Domain

        story = _story_service(request).fetch_story(story_id)

        content_type = request.headers.get("Accept")
        if content_type == "text/vnd.graphviz":
            project_id = utils.default_arg(request, "project_id", config.project_name)
            domain_dict = _project_service(request).get_domain(project_id)
            domain = Domain.from_dict(domain_dict)

            visualization = await _story_service(request).visualize_stories(
                [story], domain
            )

            if visualization:
                return response.text(visualization)
            else:
                return error(
                    406,
                    "VisualizationNotAvailable",
                    "Cannot produce a visualization for the requested story",
                )
        else:
            if story:
                return response.json(story)

        return error(
            404, "StoryNotFound", "Story for id {} could not be found".format(story_id)
        )

    @stack_endpoints.route("/evaluate", methods=["POST"])
    @platform_scoped("allEvaluations.create")
    async def update_conversation_evaluations(request):
        event_service = _event_service(request)
        conversation_ids = event_service.sender_ids()
        for _id in conversation_ids:
            story = event_service.story_for_sender_id(_id)
            content, status = await _stack_service(request, "worker").evaluate_story(
                story
            )

            if status != 200:
                return error(
                    status,
                    "CoreEvaluationFailed",
                    "Failed to create evaluation for conversation_id "
                    "'{}'.".format(_id),
                    details=content,
                )
            try:
                event_service.update_evaluation(_id, content)
            except ValueError as e:
                return error(
                    status,
                    "StoringEvaluationFailed",
                    "Failed to store evaluation for conversation_id "
                    "'{}'.".format(_id),
                    details=e,
                )

        return response.text("", 204)

    @stack_endpoints.route("/evaluations", methods=["GET", "HEAD"])
    @platform_scoped("allEvaluations.list")
    async def get_evaluations(request):
        event_service = _event_service(request)
        conversation_ids = event_service.sender_ids()
        evaluations = []
        for _id in conversation_ids:
            _eval = event_service.get_evaluation(_id)
            evaluations.append({_id: _eval})
        return response.json(evaluations, headers={"X-Total-Count": len(evaluations)})

    @stack_endpoints.route(
        "/conversations/<conversation_id>/evaluation", methods=["GET", "HEAD"]
    )
    @platform_scoped("clientEvaluation.get")
    async def get_evaluation(request, conversation_id):
        evaluation = _event_service(request).get_evaluation(conversation_id)
        return response.json(evaluation)

    @stack_endpoints.route(
        "/conversations/<conversation_id>/evaluation", methods=["DELETE"]
    )
    @platform_scoped("clientEvaluation.delete")
    async def delete_evaluation(request, conversation_id):
        _event_service(request).delete_evaluation(conversation_id)
        return response.text("", 204)

    @stack_endpoints.route(
        "/conversations/<conversation_id>/evaluation", methods=["PUT"]
    )
    @platform_scoped("clientEvaluation.update")
    async def put_evaluation(request, conversation_id):
        event_service = _event_service(request)
        story = event_service.story_for_sender_id(conversation_id)

        if not story:
            return error(
                404,
                "ClientNotFound",
                "Client for conversation_id {} could not be found"
                "".format(conversation_id),
            )

        content, status = await _stack_service(request, "worker").evaluate_story(story)
        if status != 200:
            return error(
                status,
                "CoreEvaluationFailed",
                "Failed to create evaluation for conversation_id "
                "{}".format(conversation_id),
                details=content,
            )

        try:
            event_service.update_evaluation(conversation_id, content)

            return response.json(content)
        except ValueError as e:
            return error(
                status,
                "StoringEvaluationFailed",
                "Failed to store evaluation for conversation_id "
                "'{}'.".format(conversation_id),
                details=e,
            )

    @stack_endpoints.route("/conversations/<conversation_id>/predict", methods=["POST"])
    @platform_scoped("actionPrediction.get")
    async def predict_next_action(request, conversation_id):
        included_events = utils.default_arg(request, "include_events", "ALL")

        content, status = await _stack_service(
            request, "production"
        ).predict_next_action(conversation_id, included_events)

        if status != 200:
            return error(
                status,
                "ActionPredictionFailed",
                "Failed to predict the next action for "
                "conversation_id '{}'. Stack returned status '{}' with "
                "error message '{}'.".format(conversation_id, status, content),
            )

        return response.json(content, status)

    @stack_endpoints.route(
        "/conversations/<conversation_id>/messages/<message_timestamp>/flag",
        methods=["PUT"],
    )
    @platform_scoped("messageFlags.update")
    async def flag_message(request, conversation_id, message_timestamp):
        try:
            message = float(message_timestamp)
            _event_service(request).add_flagged_message(conversation_id, message)
            return response.text("", 201)
        except Exception as e:
            return error(400, "Failed to flag message", e)

    @stack_endpoints.route(
        "/conversations/<conversation_id>/messages/<message_timestamp>/flag",
        methods=["DELETE"],
    )
    @platform_scoped("messageFlags.delete")
    async def delete_flag_from_conversation(
        request, conversation_id, message_timestamp
    ):
        try:
            message = float(message_timestamp)
            _event_service(request).delete_flagged_message(conversation_id, message)
            return response.text("", 200)
        except Exception as e:
            return error(400, "Failed to delete message flag", e)

    @stack_endpoints.route(
        "/conversations/<conversation_id>/messages/<message_timestamp>/intent",
        methods=["PUT"],
    )
    @platform_scoped("messageIntents.update")
    @inject_rasa_x_user()
    async def correct_message(request, conversation_id, message_timestamp, user):
        try:
            message = float(message_timestamp)
            intent = request.json
            if not utils.check_schema("intent", intent):
                return error(
                    400, "WrongSchema", "Please check the schema of your intent payload"
                )

            project_id = utils.default_arg(request, "project_id", config.project_name)
            _event_service(request).correct_message(
                conversation_id, message, intent, user, project_id
            )

            return response.text("", 200)
        except Exception as e:
            return error(400, "MessageRelabellingError.", e)

    @stack_endpoints.route(
        "/conversations/<conversation_id>/messages/<message_timestamp>/intent",
        methods=["DELETE"],
    )
    @platform_scoped("messageIntents.delete")
    async def delete_message_correction(request, conversation_id, message_timestamp):
        try:
            message = float(message_timestamp)

            project_id = utils.default_arg(request, "project_id", config.project_name)
            _event_service(request).delete_message_correction(
                conversation_id, message, project_id
            )
            return response.text("", 200)
        except Exception as e:
            return error(400, "MessageRelabellingError.", e)

    @stack_endpoints.route("/stories", methods=["POST"])
    @platform_scoped("stories.create")
    @inject_rasa_x_user(allow_api_token=True)
    async def add_stories(request, user=None):
        story_string = convert_bytes_to_string(request.body)
        try:
            saved_stories = await _story_service(request).save_stories(
                story_string, user["team"], username=user["username"]
            )
            return response.json(
                saved_stories, headers={"X-Total-Count": len(saved_stories)}
            )
        except StoryParseError as e:
            logger.error(e.message)
            return error(
                400, "StoryParseError", "Failed to parse story.", details=e.message
            )

    @stack_endpoints.route("/stories", methods=["PUT"])
    @platform_scoped("bulkStories.update")
    @inject_rasa_x_user()
    async def add_bulk_stories(request, user=None):
        story_string = convert_bytes_to_string(request.body)
        try:
            saved_stories = await _story_service(request).replace_stories(
                story_string, user["team"], user["username"]
            )
            if saved_stories:
                return response.json(
                    saved_stories, headers={"X-Total-Count": len(saved_stories)}
                )
        except StoryParseError as e:
            logger.error(e.message)
            return error(
                400, "StoryParseError", "Failed to parse stories.", details=e.message
            )

    @stack_endpoints.route("/stories.md", methods=["GET", "HEAD"])
    @platform_scoped("bulkStories.get", allow_api_token=True)
    async def get_bulk_stories(request):
        request.headers["Accept"] = "text/markdown"
        return await get_stories(request)

    @stack_endpoints.route("/stories/<story_id>", methods=["PUT"])
    @platform_scoped("stories.update")
    @inject_rasa_x_user()
    async def modify_story(request, story_id, user=None):
        story_string = convert_bytes_to_string(request.body)
        try:
            updated_story = await _story_service(request).update_story(
                story_id, story_string, user
            )
            if not updated_story:
                return response.text("Story could not be found", 404)
            else:
                return response.json(updated_story)
        except StoryParseError as e:
            logger.error(e.message)
            return error(
                400, "StoryParseError", "Failed to modify story.", details=e.message
            )

    @stack_endpoints.route("/stories/<story_id>", methods=["DELETE"])
    @platform_scoped("stories.delete")
    async def delete_story(request, story_id):
        deleted = _story_service(request).delete_story(story_id)
        if deleted:
            return response.text("", 204)
        return error(
            404,
            "StoryNotFound",
            "Failed to delete story with story_id {}.".format(story_id),
        )

    @stack_endpoints.route("/chat", methods=["POST"])
    @platform_scoped("messages.create")
    async def chat_to_bot(request):
        message_object = request.json
        if not utils.check_schema("message", message_object):
            return error(
                400, "WrongSchema", "Please check the schema of your message payload."
            )
        environment = utils.deployment_environment_from_request(request, "production")
        service = _stack_service(request, "production")
        if not service:
            return error(
                404,
                "ServiceNotFound",
                "Service for requested environment '{}' not found."
                "".format(environment),
            )
        message_response = await service.send_message(
            message_object, token=request.headers.get("Authorization")
        )
        if message_response is not None:
            return response.json(message_response)
        else:
            return error(
                404,
                "MessageSendFailed",
                "Message '{}' could not be sent to environment '{}'"
                "".format(message_object["message"], environment),
            )

    @stack_endpoints.route("/auth/jwt", methods=["POST"])
    async def issue_signed_jwt(request):
        id_object = request.json
        if not utils.check_schema("signed_jwt", id_object):
            return error(
                400,
                "WrongSchema",
                "Please check the schema of your chat token payload.",
            )

        chat_token = _project_service(request).get_token()
        if id_object["chat_token"] != chat_token.token:
            return error(401, "InvalidChatToken", "Chat token is not valid.")

        if _project_service(request).has_token_expired(id_object["chat_token"]):
            return error(401, "ChatTokenExpired", "Chat token has expired.")

        # set JWT expiration to one day from current time or chat token
        # expiration time, whichever is lower
        expires = min(int(time.time()) + 24 * 60 * 60, chat_token.expires)

        jwt_object = {
            "username": id_object["conversation_id"],
            "exp": expires,
            "user": {"username": id_object["conversation_id"], "roles": [GUEST_ROLE]},
            "scopes": normalise_permissions(guest_permissions()),
        }

        encoded = jwt.encode(jwt_object, config.jwt_secret, algorithm="HS256").decode(
            "utf-8"
        )

        return response.json({"access_token": encoded})

    return stack_endpoints
