import asyncio
import logging
import os
import typing
import webbrowser
from typing import Text, Tuple, Generator, Optional, Dict, List, Set, Any

from rasa.cli.utils import print_success
from rasa.nlu.training_data import TrainingData
from rasa.nlu.training_data.formats import RasaReader
from rasa.nlu.training_data.loading import _guess_format
from rasa.nlu.utils import write_to_file
from rasa.utils.io import read_yaml_file, read_file
from sanic import Sanic
from sqlalchemy.orm import Session

from rasax.community import config
from rasax.community.services.role_service import RoleService
from rasax.community.services.user_service import UserService
from rasax.community.utils import random_password

if typing.TYPE_CHECKING:
    from rasax.community.services.settings_service import SettingsService

logger = logging.getLogger(__name__)


def inject_config(config_path: Text, settings_service: "SettingsService"):
    """Load a configuration file from `path` and save it to the database."""

    if not os.path.exists(config_path):
        logger.warning(
            "Failed to inject Rasa configuration. The file '{}' "
            "does not exists."
            "".format(os.path.abspath(config_path))
        )
        return

    yaml_config = read_yaml_file(config_path)
    if not yaml_config:
        logger.warning(
            "Failed to inject Rasa configuration:\n"
            "Reading of yaml '{}' file failed. Most likely the "
            "file was not found or uses the wrong syntax."
            "".format(os.path.abspath(config_path))
        )

    stack_config = settings_service.save_config(
        config.team_name, "default", yaml_config, config_path
    )
    if stack_config:
        logger.debug(
            "Loaded local configuration from '{}' into "
            "database".format(os.path.abspath(config_path))
        )
    else:
        logger.warning(
            "Failed to inject Rasa configuration. The file '{}'"
            "was found but could not be saved to the database."
            "".format(os.path.abspath(config_path))
        )


def _read_data(paths: List[Text]) -> Generator[Tuple[Text, Text], None, None]:
    for filename in paths:
        yield read_file(filename), filename


def inject_nlu_data(
    nlu_files: Set[Text], project_id: Text, username: Text, data_service: "DataService"
) -> TrainingData:
    """Load Rasa NLU training data from `path` and save it to the database."""

    # delete existing data in db if files are provided
    if nlu_files:
        data_service.delete_data()

    training_data = TrainingData()

    for data, path in _read_data(list(nlu_files)):
        logger.debug("Injecting NLU data from file '{}' to database.".format(path))

        data_format = _guess_format(path)
        if data_format not in {"md", "json"}:
            logger.debug(
                "Invalid data format for file '{}'. "
                "Accepting 'md' and 'json' but found "
                "'{}'.".format(path, data_format)
            )
            continue

        additional_data = data_service.save_bulk_data(
            project_id,
            data,
            data_format,
            username=username,
            filename=path,
            dump_data_in_local_mode=False,
        )

        if additional_data:
            training_data = training_data.merge(additional_data)

    logger.debug(
        "Injected {} NLU training data examples."
        "".format(len(training_data.training_examples))
    )
    return training_data


def dump_nlu_data(
    data_service: "DataService", project: Text = config.project_name
) -> None:
    """Dump Rasa NLU data in database to file at `path`."""

    data_filenames = data_service.get_all_filenames(project)
    for fname in data_filenames:
        path = os.path.join(fname)
        logger.debug("Dumping NLU data to file '{}'.".format(path))

        nlu_format = data_service.create_formatted_training_data(fname, project)

        training_data = RasaReader().read_from_json(nlu_format)
        write_to_file(path, training_data.as_markdown())


async def inject_stories(
    story_files: Set[Text],
    story_service: "StoryService",
    username: Text,
    team: Text = config.team_name,
) -> List[Dict[Text, Any]]:
    """Load Core stories from `data_directory` and save to the database."""

    # delete existing data in db if files are provided
    if story_files:
        await story_service.delete_stories(team)

    story_blocks = []
    additional_blocks = []

    for data, path in _read_data(list(story_files)):
        logger.debug("Injecting stories from file '{}' to database.".format(path))
        additional_blocks = await story_service.save_stories(
            data, team, path, username, dump_stories_in_local_mode=False
        )
        story_blocks.extend(additional_blocks)

    logger.debug("Injected {} Core stories.".format(len(additional_blocks)))
    return story_blocks


def inject_domain(
    domain_path: Text,
    projects_service: "ProjectsService",
    project_id: Text = config.project_name,
    username: Text = config.default_username,
) -> Dict[Text, Any]:
    """Load Rasa Core domain at `path` and save it to database."""

    if not os.path.exists(domain_path):
        logger.warning(
            "Failed to inject domain file. The file '{}' "
            "does not exists."
            "".format(os.path.abspath(domain_path))
        )
        return {}

    data = read_yaml_file(domain_path)

    # inject domain including templates
    projects_service.store_domain(
        data, project_id, path=domain_path, store_templates=True, username=username
    )

    return data


def inject_project(
    settings_service: "SettingsService",
    user_service: "UserService",
    project_id: Text,
    team: Text,
    is_in_local_mode: bool,
    user: Optional[Dict] = None,
) -> None:
    """Create project and settings."""

    existing = settings_service.get(team, project_id)
    if existing is None:
        if user is not None:
            user_service.create_project(user, project_id)
        settings_service.init_project(
            team, project_id, is_in_local_mode=is_in_local_mode
        )
        logger.debug("Created project '{}'.".format(project_id))
    else:
        logger.debug("Found project '{}'.".format(project_id))


def create_project_and_settings(
    _settings_service: "SettingsService", _role_service: "RoleService", team: Text
) -> None:
    project_id = config.project_name

    existing = _settings_service.get(team, project_id)

    if existing is None:
        _settings_service.init_project(team, project_id)

    _role_service.init_roles(project_id=project_id)


def create_community_user(session: Session, app: Sanic) -> None:
    from rasax.community.local import (
        LOCAL_USERNAME,
        LOCAL_PROJECT_NAME,
        LOCAL_TEAM_NAME,
    )
    from rasax.community.services.settings_service import SettingsService

    role_service = RoleService(session)
    role_service.init_roles(project_id=LOCAL_PROJECT_NAME)

    settings_service = SettingsService(session)
    password = _initialize_local_password(settings_service)

    user_service = UserService(session)
    user_service.insert_or_update_user(LOCAL_USERNAME, password, LOCAL_TEAM_NAME)

    app.add_task(_startup_info)


def _initialize_local_password(settings_service: "SettingsService") -> Text:
    """Update Rasa X password.

    Check if `RASA_X_PASSWORD` environment variable is set, and use it as password.
    Generate new password if it is not set.
    """

    password = os.environ.get("RASA_X_PASSWORD", random_password())

    # update password in config
    config.rasa_x_password = password

    # commit password to db
    settings_service.save_local_password(password)

    return password


async def _startup_info() -> None:
    """Open a browser window pointing to the UI."""

    from rasax.community.local import LOCAL_USERNAME

    startup_info_shown_env = "RASA_STARTUP_INFO_SHOWN"
    startup_info_shown = os.environ.get(startup_info_shown_env)
    if startup_info_shown:
        return

    os.environ[startup_info_shown_env] = True

    if config.rasa_x_hostname:
        server_url = config.rasa_x_hostname
    else:
        server_url = "http://localhost:{}".format(config.self_port)

    login_url = "{}/login?username={}&password={}" "".format(
        server_url, LOCAL_USERNAME, config.rasa_x_password
    )

    print_success("\nThe server is running at {}\n".format(login_url))
    webbrowser.open_new_tab(login_url)

    # Wait a bit to make sure all workers have tested against this
    await asyncio.sleep(2)
    # Unset env variable so it's shown again at the next start
    del os.environ[startup_info_shown_env]
