import time
from typing import Any, Text, Dict

from rasax.community.services.story_service import StoryService

TEST_STORY_NAME_1 = "Test Story"
TEST_STORY_NAME_2 = "Test Story 2"
TEST_FILENAME = "test story filename"

TEST_STORY = """

## {}

* intent1
  - action1
  - action2
* intent2
  - action1

## {}
* intent_other
  - action_other

""".format(
    TEST_STORY_NAME_1, TEST_STORY_NAME_2
)


async def test_save_story(story_service: StoryService, user: Dict[Text, Any]):
    current_time = time.time()
    await story_service.save_stories(
        TEST_STORY,
        user["team"],
        TEST_FILENAME,
        username=user["username"],
        dump_stories_in_local_mode=False,
    )

    actual = story_service.fetch_stories()

    assert len(actual) == 2
    assert actual[0]["name"] == TEST_STORY_NAME_1
    assert actual[1]["name"] == TEST_STORY_NAME_2

    for story in actual:
        assert story["story"] in TEST_STORY
        assert story["annotation"]["user"] == user["username"]
        assert story["annotation"]["time"] > current_time


def test_get_story_with_text_query(story_service: StoryService):
    query = "other"
    actual = story_service.fetch_stories(query)

    assert len(actual) == 1
    assert actual[0]["name"] == TEST_STORY_NAME_2


def test_get_story_with_filename(story_service: StoryService):
    actual = story_service.fetch_stories(filename=TEST_FILENAME)

    assert len(actual) == 2


def test_get_story_with_invalid_text_query(story_service: StoryService):
    query = "lorem ipsum"
    actual = story_service.fetch_stories(query)

    assert not actual


def test_get_stories_with_fields(story_service: StoryService, user: Dict[Text, Any]):
    fields = [("id", True), ("annotation.user", True)]

    actual = story_service.fetch_stories(field_query=fields)

    assert len(actual) == 2
    assert actual[0]["id"] == 1
    assert actual[1]["id"] == 2

    assert all(story["annotation"]["user"] == user["username"] for story in actual)


async def test_update_story(story_service: StoryService, user: Dict[Text, Any]):
    other_story_name = "other story"
    other_story = """
    ## {}
    * intent
      - action
    """.format(
        other_story_name
    )

    story_service._dump_stories_in_local_mode = lambda: True

    await story_service.update_story("1", other_story, user)
    actual = story_service.fetch_stories()

    assert len(actual) == 2
    assert actual[0]["story"] in other_story
    assert actual[0]["name"] == other_story_name


async def test_replace_stories(story_service: StoryService, user: Dict[Text, Any]):
    replacement_story = """
    ## replacement
    * intent
      - action
    """
    await story_service.replace_stories(
        replacement_story, user["team"], TEST_FILENAME, username=user["username"]
    )

    actual = story_service.fetch_stories()

    assert len(actual) == 1
    assert actual[0]["name"] == "replacement"


def test_get_filenames(story_service: StoryService, user: Dict[Text, Any]):
    assert story_service.get_filenames(user["team"]) == [TEST_FILENAME]


def test_delete_story(story_service: StoryService):
    assert story_service.delete_story("1")
    assert not story_service.fetch_stories()
